@@ -33,6 +33,7 @@ import javax.swing.event.*;
 import javax.accessibility.AccessibleContext;
 import javax.accessibility.AccessibleRole;
 import javax.accessibility.AccessibleText;
+import javax.accessibility.AccessibleEditableText;
 
 public class JavaBridge {
 
@@ -42,6 +43,7 @@ public class JavaBridge {
  static java.util.Hashtable accessibleObjectTable = new java.util.Hashtable();
  static EventQueue accessQueue = null;
  static AccessibleContext oldFocusSource = null;
+ static String oldFocusText = "";
  static AccessibleContext oldFocusCtx = null;
  static AccessibleContext savedFocusSource = null;
  static java.lang.Object oldEventSource = null;
@@ -56,9 +58,9 @@ public class JavaBridge {
 
 	static {
  	  globalPropertyChangeListener = new java.beans.PropertyChangeListener() {
-		public void propertyChange (java.beans.PropertyChangeEvent e) {
-			// System.out.println ("Prop Change Event: " + e.getSource()+" "+e.getPropertyName());
-			String propertyName = e.getPropertyName();
+		  public void propertyChange (java.beans.PropertyChangeEvent e) {
+		        String propertyName = e.getPropertyName();
+//			System.out.println ("Prop Change Event: " + e.getSource()+" "+e.getPropertyName());
 			if (propertyName == AccessibleContext.ACCESSIBLE_CARET_PROPERTY)
 				dispatchCaretProperty(e);
 			else if (propertyName == AccessibleContext.ACCESSIBLE_TEXT_PROPERTY)
@@ -96,22 +98,22 @@ public class JavaBridge {
 				if (source instanceof AccessibleContext ) {
 					role = ((AccessibleContext) source).getAccessibleRole();
 					if ( role == javax.accessibility.AccessibleRole.TEXT ) 
-                        textEvent = true;
+					    textEvent = true;
 					else if ( role == javax.accessibility.AccessibleRole.MENU_BAR ) 
-                        dispatchFocusEvent (source);
-                    else if ( role == javax.accessibility.AccessibleRole.PAGE_TAB_LIST ) {
-                        // Generate a focus event for the tab selected.
-                        // Because TABs are not actually components in Java
-                        // they don't generate focus, so we need to simulate.
-                        javax.accessibility.AccessibleSelection aSelection = 
-                            ((AccessibleContext) source).getAccessibleSelection();
-                        if ( aSelection != null && 
-                             aSelection.getAccessibleSelectionCount() > 0 ) {
-                            java.lang.Object child = aSelection.getAccessibleSelection(0);
-                            dispatchFocusEvent (child);
-                        }
-                    }
-                }
+					    dispatchFocusEvent (source);
+					else if ( role == javax.accessibility.AccessibleRole.PAGE_TAB_LIST ) {
+					    // Generate a focus event for the tab selected.
+					    // Because TABs are not actually components in Java
+					    // they don't generate focus, so we need to simulate.
+					    javax.accessibility.AccessibleSelection aSelection = 
+						((AccessibleContext) source).getAccessibleSelection();
+					    if ( aSelection != null && 
+						 aSelection.getAccessibleSelectionCount() > 0 ) {
+						java.lang.Object child = aSelection.getAccessibleSelection(0);
+						dispatchFocusEvent (child);
+					    }
+					}
+				}
 //				else if (source instanceof AccessibleContext) 
 //					System.out.println ("Selection source role: " +
 //                                         ((AccessibleContext) source).getAccessibleRole());
@@ -120,16 +122,16 @@ public class JavaBridge {
 //				          ((javax.accessibility.Accessible) source).
 //							    getAccessibleContext().
 //							    getAccessibleRole());
-                if ( textEvent ) {
-                    dispatchEvent(source,
-                              "object:text-selection-changed", 
-                              0, 0);
-                }
-                else {
-                    dispatchEvent(source,
-                              "object:selection-changed", 
-                              0, 0);
-                }
+				if ( textEvent ) {
+				    dispatchEvent(source,
+						  "object:text-selection-changed", 
+						  0, 0);
+				}
+				else {
+				    dispatchEvent(source,
+						  "object:selection-changed", 
+						  0, 0);
+				}
 			}
 			else if (propertyName == 
 				 AccessibleContext.ACCESSIBLE_VISIBLE_DATA_PROPERTY)
@@ -198,67 +200,69 @@ public class JavaBridge {
 					      "object:property-change:accessible-table-row-description",
 					      0, 0);
 			else if (propertyName == AccessibleContext.ACCESSIBLE_STATE_PROPERTY) {
-				java.lang.Object value = e.getNewValue();
-                    		java.lang.Object source = e.getSource();
-				if (value == javax.accessibility.AccessibleState.ARMED ) {
-                    javax.accessibility.AccessibleRole role = null;
-                    javax.accessibility.Accessible parent = null;
-                    javax.accessibility.AccessibleRole parent_role = null;
+			    java.lang.Object value = e.getNewValue();
+			    java.lang.Object source = e.getSource();
+			    if (value == javax.accessibility.AccessibleState.ARMED ) {
+				javax.accessibility.AccessibleRole role = null;
+				javax.accessibility.Accessible parent = null;
+				javax.accessibility.AccessibleRole parent_role = null;
+				
+				if (source instanceof AccessibleContext ) {
+				    parent = ((AccessibleContext) source).getAccessibleParent();
+				    if ( parent != null )
+					parent_role = parent.getAccessibleContext().getAccessibleRole();
+				    role = ((AccessibleContext) source).getAccessibleRole();
+				}
+				else if (source instanceof javax.accessibility.Accessible) {
+				    parent = ((javax.accessibility.Accessible) source).
+					getAccessibleContext().getAccessibleParent();
+				    if ( parent != null )
+					parent_role = parent.getAccessibleContext().getAccessibleRole();
+				    role = ((javax.accessibility.Accessible) source).
+					getAccessibleContext().getAccessibleRole();
+				}
+				
+//				if ( parent_role != null )
+//				    System.err.println ("State changed source parent_role " + parent_role );
+//				else 
+//				    System.err.println ("State changed source parent_role null");
+//				if ( role != null )
+//				    System.err.println ("State changed source role " + role );
+//				else 
+//				    System.err.println ("State changed source role null");
+				
+				if ( role != null ) {
+				    if ( role == javax.accessibility.AccessibleRole.MENU_ITEM ||
+					 role == javax.accessibility.AccessibleRole.MENU ) {
+					
+					dispatchFocusEvent (source);
+				    }
+				    else if ( parent_role != null && 
+					      ( parent_role == javax.accessibility.AccessibleRole.MENU && 
+						( role == javax.accessibility.AccessibleRole.CHECK_BOX ||
+						  role == javax.accessibility.AccessibleRole.RADIO_BUTTON ) 
+						  ) ) {
+					dispatchFocusEvent (source);
+				    }
+				}
+			    } else if (value == javax.accessibility.AccessibleState.SELECTED
+				       && source instanceof AccessibleContext ) {
+				javax.accessibility.AccessibleRole role = null;
+				role = ((AccessibleContext) source).getAccessibleRole ();
+				if ( role == javax.accessibility.AccessibleRole.MENU ) {
+				    dispatchFocusEvent (source);
+				}
+			    } else if (value == javax.accessibility.AccessibleState.FOCUSED) {
+				dispatchFocusEvent (source);
+			    }
 
-                    if (source instanceof AccessibleContext ) {
-                        parent = ((AccessibleContext) source).getAccessibleParent();
-                        if ( parent != null )
-                            parent_role = parent.getAccessibleContext().getAccessibleRole();
-                        role = ((AccessibleContext) source).getAccessibleRole();
-                    }
-                    else if (source instanceof javax.accessibility.Accessible) {
-                        parent = ((javax.accessibility.Accessible) source).
-                                    getAccessibleContext().getAccessibleParent();
-                        if ( parent != null )
-                            parent_role = parent.getAccessibleContext().getAccessibleRole();
-                        role = ((javax.accessibility.Accessible) source).
-                                    getAccessibleContext().getAccessibleRole();
-                    }
-                    
-//                    if ( parent_role != null )
-//                        System.err.println ("State changed source parent_role " + parent_role );
-//                    else 
-//                        System.err.println ("State changed source parent_role null");
-//                    if ( role != null )
-//                        System.err.println ("State changed source role " + role );
-//                    else 
-//                        System.err.println ("State changed source role null");
-
-                    if ( role != null ) {
-				        if ( role == javax.accessibility.AccessibleRole.MENU_ITEM ||
-				             role == javax.accessibility.AccessibleRole.MENU ) {
-
-                            dispatchFocusEvent (source);
-                        }
-                        else if ( parent_role != null && 
-                                  ( parent_role == javax.accessibility.AccessibleRole.MENU && 
-                                   ( role == javax.accessibility.AccessibleRole.CHECK_BOX ||
-                                     role == javax.accessibility.AccessibleRole.RADIO_BUTTON ) 
-                                  ) ) {
-                            dispatchFocusEvent (source);
-                        }
-                    }
-				} else if (value == javax.accessibility.AccessibleState.SELECTED
-					&& source instanceof AccessibleContext ) {
-                    javax.accessibility.AccessibleRole role = null;
-					role = ((AccessibleContext) source).getAccessibleRole ();
-					if ( role == javax.accessibility.AccessibleRole.MENU ) {
-                        dispatchFocusEvent (source);
-                    }
-				} else if (value == javax.accessibility.AccessibleState.FOCUSED) {
-                        		dispatchFocusEvent (source);
-                    }
-
-				dispatchStateProperty(e);
+			    // System.err.println ("State changed: " + source + ": new value : " + value);				
+			    
+			    dispatchStateProperty(e);
 			}
-		}	
-	  };
-		
+		  }	
+	      };
+	  
 	  focusListener = new java.awt.event.FocusListener() {
             public void focusGained(FocusEvent e) {
                 dispatchFocusEvent(e.getSource());
@@ -339,6 +343,7 @@ public class JavaBridge {
 					  int detail2,
                                           org.omg.CORBA.Any any) {
 		AccessibleImpl source;
+		// System.out.println ("dispatching event with ANY type " + any.type ());
 		try {
 			AccessibleContext ac = null;
 			if (eventSource instanceof AccessibleContext) 
@@ -444,6 +449,12 @@ public class JavaBridge {
 				savedFocusSource = ctx;
 			}
 			oldFocusSource = ctx;
+			if ((ctx != null) && (ctx.getAccessibleText () != null)) {
+			    AccessibleText axt = ctx.getAccessibleText ();
+			    oldFocusText = getTextRange (axt, 0, 
+							 axt.getCharCount ());
+			}
+			else oldFocusText = "";
 		}
 
             //
@@ -501,6 +512,41 @@ public class JavaBridge {
         lastTextSizes.put(ac, new Integer(size));
     }
 
+
+    // cut-and-pasted from TextImpl.java
+    private static String getTextRange ( AccessibleText accText, int startOffset, int endOffset) {
+	String text = "";
+	int offset = startOffset;
+	String word = accText.getAtIndex (
+	    javax.accessibility.AccessibleText.WORD, 
+	    startOffset);
+	// go char-by-char until word changes,
+	while (offset < endOffset && word.equals (accText.getAtIndex (
+						      javax.accessibility.AccessibleText.WORD, offset) ) ) {
+	    String character = accText.getAtIndex (
+		javax.accessibility.AccessibleText.CHARACTER, 
+		offset);
+	    ++offset;
+	    if (character != null)
+		text += character; 
+	    // TODO: charbuff would be faster here
+	}
+	// then word by word until offset crosses
+	while (offset < endOffset) {
+	    word = accText.getAtIndex (
+		javax.accessibility.AccessibleText.WORD, 
+		offset);
+	    offset += word.length();
+	    if (offset > endOffset) 
+		text += word.substring (0,
+					word.length() - (offset - endOffset));
+	    else
+		text += word;
+	} // TODO: sentence-by-sentence for even longer spans.
+
+	return text;
+    }
+
     //
     // In order to overcome some deficiencies in the JAA a private interface
     // was agreed with SO/OOo to handle more detailed information about text
@@ -545,6 +591,7 @@ public class JavaBridge {
                          .getAccessibleContext();
 
                 AccessibleText aText = ac.getAccessibleText();
+		AccessibleEditableText aeText = ac.getAccessibleEditableText();
 
                 int charCount = aText.getCharCount();
                 Integer prevCount = (Integer)lastTextSizes.put(ac, new Integer(charCount));
@@ -557,13 +604,48 @@ public class JavaBridge {
                 if (charCount > prevCharCount) {
                     eventString += ":insert";
                     oldValue = charCount - prevCharCount;
-                }
+		    if (aeText != null) {
+			any.insert_string (aeText.getTextRange (((Integer)newValueObject).intValue(),
+							       ((Integer)newValueObject).intValue() + oldValue));
+			if (ac == oldFocusSource)
+			    oldFocusText = aeText.getTextRange (0, charCount);
+		    }
+		    else if ((ac != null) && (ac == oldFocusSource)) {
+			any.insert_string (oldFocusText.substring (
+				       ((Integer)newValueObject).intValue(),
+				       ((Integer)newValueObject).intValue() + 
+				       oldValue));
+			oldFocusText = getTextRange (aText, 0, charCount);
+		    }
+		}
                 else if ( charCount < prevCharCount ) {
                     eventString += ":delete";
                     oldValue = prevCharCount - charCount;
+		    
+		    if ((ac != null) && (ac == oldFocusSource) && (oldFocusText != "")) {
+			int substart = ((Integer)newValueObject).intValue();
+			int subend = substart + oldValue;
+			if (oldFocusText.length() < subend)
+			    System.err.println ("WARNING: subend " + subend + " > len : " + oldFocusText + " - substart = " + substart);
+
+			any.insert_string (oldFocusText.substring (
+				       ((Integer)newValueObject).intValue(),
+				       ((Integer)newValueObject).intValue() + 
+				       oldValue));
+		    }
+		    // otherwise, can't retrieve the lost contents
+		    oldFocusText = getTextRange (aText, 0, charCount);
                 }
                 else {
                     eventString += ":replace";
+		    if (aeText != null) {
+			String replacedString = aeText.getTextRange (0, charCount);
+			any.insert_string (replacedString);
+			if (ac == oldFocusSource)
+			    oldFocusText = replacedString;
+		    }
+		    else if (ac == oldFocusSource)
+			oldFocusText = getTextRange (aText, 0, charCount);
                 }
                 newValue = ((Integer)newValueObject).intValue();
             }
@@ -600,7 +682,7 @@ public class JavaBridge {
                     newValue = oldInts[0].intValue();
                     oldValue = oldInts[1].intValue();
                     dispatchEvent(e.getSource(), eventString + ":delete", 
-                                  newValue, oldValue);
+                                  newValue, oldValue, any);
 
                     // Now setup so that an insert will be sent below.
                     eventString += ":insert";
@@ -645,7 +727,7 @@ public class JavaBridge {
                     // Delete
                     eventString += ":delete";
                     newValue = oldInts[0].intValue();
-                    oldValue = oldInts[1].intValue();
+                    oldValue = oldInts[1].intValue() - oldInts[0].intValue();
                     text = (String)oldObjs[2];
                 }
                 else if ( newValueObject != null && oldValueObject != null ) {
@@ -721,10 +803,11 @@ public class JavaBridge {
                 }
             }
         }
-		dispatchEvent(e.getSource(),
-			      eventName,
-			      0,
-			      0);
+	//System.err.println ("dispatching child event: "+ eventName);
+	dispatchEvent(e.getSource(),
+		      eventName,
+		      0,
+		      0);
 	}
 
 	public static void dispatchStateProperty(java.beans.PropertyChangeEvent e) {
@@ -745,16 +828,18 @@ public class JavaBridge {
 			stateType = new StateTypeAdapter((javax.accessibility.AccessibleState)oldValue);
             detail1 = 0;
 		}
-        // State COLLAPSED has been deprecated in AT-SPI, should now be using
-        // the state EXPANDED wit the value of 0 instead.
-        if ( stateType.value() == StateType._STATE_COLLAPSED ) {
-            stateType = 
-                new StateTypeAdapter(javax.accessibility.AccessibleState.EXPANDED);
-            detail1 = 0;
-        }
+		// State COLLAPSED has been deprecated in AT-SPI, should now be using
+		// the state EXPANDED wit the value of 0 instead.
+		if ( stateType.value() == StateType._STATE_COLLAPSED ) {
+		    stateType = 
+			new StateTypeAdapter(javax.accessibility.AccessibleState.EXPANDED);
+		    detail1 = 0;
+		}
 
 		String eventName = "object:state-changed:" + stateType.getName();
 
+		//System.out.println ("dispatching state event: " + eventName);
+
 		if (object == oldEventSource &&
 		    eventName.compareTo(oldEventName) == 0 &&
 		    detail1 == oldDetail1)
@@ -764,7 +849,7 @@ public class JavaBridge {
 		oldEventName = eventName;
 		oldDetail1 = detail1;
 
-		dispatchEvent(object, eventName, detail1, /*detail2*/ 0  );
+		dispatchEvent(e.getSource(), eventName, detail1, /*detail2*/ 0  );
 	}
 
     static private class AttachListenerRunnable implements Runnable {
