@@ -36,8 +36,6 @@ public class TextImpl extends UnknownImpl implements TextOperations {
     // Until JAA implements directly, we've agreed with SO/OOo to pass the
     // value through.
     static final int LINE_PART = javax.accessibility.AccessibleText.SENTENCE + 1;
-    // XXX Not yet part of GNOME, but will be...
-    // static final int RUN_PART = LINE_PART + 1;
 
 	public class StringStruct {
 		public int startOffset;
@@ -114,6 +112,9 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                                           endOffset);
                     if (textRange != null)
                         text = textRange;
+		    // System.out.println ("Editable textRange: " + textRange);
+		    //System.out.println ("Last char: " + 
+		    //			accEditableText.getAtIndex (javax.accessibility.AccessibleText.CHARACTER, endOffset-1));
                 } catch (Exception ex) {
                     // System.err.println ("Error getting text range: " + ex);
                     ex.printStackTrace();
@@ -125,8 +126,8 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                     javax.accessibility.AccessibleText.WORD, 
                     startOffset);
                 // go char-by-char until word changes,
-                while (offset < endOffset && word == accText.getAtIndex (
-                       javax.accessibility.AccessibleText.WORD, offset) ) {
+                while (offset < endOffset && word.equals (accText.getAtIndex (
+                       javax.accessibility.AccessibleText.WORD, offset) ) ) {
                     String character = accText.getAtIndex (
                         javax.accessibility.AccessibleText.CHARACTER, 
                         offset);
@@ -137,10 +138,13 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                 }
                 // then word by word until offset crosses
                 while (offset < endOffset) {
+		    int wordlen;
                     word = accText.getAtIndex (
                         javax.accessibility.AccessibleText.WORD, 
                         offset);
-                    offset += word.length();
+		    wordlen = word.length ();
+		    // System.out.println ("word [" + word + "] len=" + wordlen);
+                    offset += wordlen;
                     if (offset > endOffset) 
                         text += word.substring (0,
                              word.length() - (offset - endOffset));
@@ -195,8 +199,8 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 
 	private StringStruct getLine (javax.accessibility.AccessibleText accText,
 				      int offset) {
-        // System.err.println("getLine ( " + offset + " )");
-		String string = null;
+	    //System.err.println("getLine ( " + offset + " )");
+		String string = "";
 		StringStruct lineStruct = new StringStruct ();
         int start = offset;
         int end = offset;
@@ -209,7 +213,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                 bounds = accText.getCharacterBounds (offset - 1);
             }
 
-            do {
+            if (start > 0) do {
                 currentBounds = accText.getCharacterBounds (--start);
             } while (start > 0 && 
                      overlapsVertically( bounds, currentBounds ));
@@ -227,7 +231,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
         }
         lineStruct.startOffset = start;
         lineStruct.endOffset = end;
-        if (string == null) 
+        if (string == "") 
             lineStruct.string = "";
         else 
             lineStruct.string = string;
@@ -241,8 +245,8 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 					int offset) {
         try {
             String string;
-//            System.err.println("concatenateText: type = " + typeToString( type) +
-//                               "; offset = " + offset );
+            //System.err.println("concatenateText: [" + text + "] type = " + typeToString( type) +
+            //                   "; offset = " + offset );
             int java_boundary_type = boundaryTypeConvert (type); 
             if (offset < 0 || offset > accText.getCharCount()) return "";
             if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_WORD_START ||
@@ -252,8 +256,8 @@ public class TextImpl extends UnknownImpl implements TextOperations {
             } else {
                 string = accText.getAfterIndex(java_boundary_type, offset);
                 if (string != null) text += string;
-                
             }
+	    //System.err.println ("Concatenated: [" + text + "]");
         } 
         catch ( Exception e ) {
             e.printStackTrace();
@@ -269,18 +273,19 @@ public class TextImpl extends UnknownImpl implements TextOperations {
             String string;
             int java_boundary_type = boundaryTypeConvert (type);
             if (offset < 0 || offset > accText.getCharCount()) return "";
-//            System.err.println("concatenateWhitespace: type = " + typeToString( type) +
-//                               "; offset = " + offset );
-            if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_WORD_START ||
-                type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_SENTENCE_START) {
+            //System.err.println("concatenateWhitespace: [" + text + "] type = " + typeToString( type) +
+            //                   "; offset = " + offset );
+            if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_WORD_START) {
                 string = accText.getAfterIndex(java_boundary_type, offset);
-                if (string != null && string.trim().length() > 0) 
+		//System.err.println ("AfterIndex string: [" + string + "]");
+		if (string != null)
                     text += string; 
-            } else {
+            } else if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_WORD_END) {
                 string = accText.getBeforeIndex(java_boundary_type, offset);
-                if (string != null && string.trim().length() > 0) 
-                    text = string + text;
-            }
+		if (string != null)
+		    text = string + text;
+            } 
+	    //System.err.println ("Concatenated: [" + text + "]");
         } 
         catch ( Exception e ) {
             e.printStackTrace();
@@ -307,41 +312,54 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 	}
 
 	private StringStruct findOffsets (javax.accessibility.AccessibleText accText,
-					  String text, int offset) {
-//        System.err.println("findOffsets ( "
-//                            + text + ", "
-//                            + offset + ") " );
-		StringStruct result = new StringStruct ();
-		String enclosingString;
-		int len = text.length ();
-		int index;
+					  String text, int offset, boolean leading_ws) {
+	    //System.err.println("findOffsets (["
+	    //		       + text + "], "
+	    //		       + offset + ") " );
+	    StringStruct result = new StringStruct ();
+	    String enclosingString;
+	    int len = text.length ();
+	    int index;
+	    
+	    // 2) try and get string from offset-len to offset+len
+	    enclosingString = this.getText (offset-len, offset+len);
+	    // System.err.println("======" + (offset-len) + "===" + (offset+len) + "===");
+	    // System.err.println("enclosingString for ["+ text + "] = " + enclosingString );
+	    // System.err.println("==========================" );
+	    
+	    // 3) find text in resulting string 
+	    index = enclosingString.indexOf (text);
+	    // System.err.println("index = " + index );
+	    // Java bug workaround; try to find string w/o newline
+	    if (index == -1) {
+		if (text.endsWith ("\n")) {
+		    // System.err.println ("FAILED TO FIND string ending in newline!");
+		    text = text.substring (0, text.length () - 1);
+		    index = enclosingString.indexOf (text);
+		}
+	    }
+	    // creep forward to add whitespace, if needed
+	    if (leading_ws) {
+		while ((index > 0) && Character.isWhitespace (enclosingString.charAt (index - 1))) 
+		    --index;
+	    }
 
-                // 2) try and get string from offset-len to offset+len
-		enclosingString = this.getText (offset-len, offset+len);
-//        System.err.println("==========================" );
-//        System.err.println("enclosingString = " + enclosingString );
-//        System.err.println("==========================" );
-
-		// 3) find text in resulting string 
-		index = enclosingString.indexOf (text);
-//        System.err.println("index = " + index );
-
-		// 4) do index math to determine where original string's offsets lie.
-		offset -= len; 
-        if ( offset < 0 ) 
-            offset = 0;
-		offset += index; 
-//        System.err.println("offset = " + offset );
-
-		// 5) return modified StringStruct
-		result.startOffset = offset;
-		result.endOffset = offset + len;
-//        System.err.println("startOffset = " + result.startOffset );
-//        System.err.println("endOffset = " + result.endOffset );
-
-		return result;
+	    // 4) do index math to determine where original string's offsets lie.
+	    offset -= len; 
+	    if ( offset < 0 ) 
+		offset = 0;
+	    offset += index; 
+	    // System.err.println("offset = " + offset );
+	    
+	    // 5) return modified StringStruct
+	    result.startOffset = offset;
+	    result.endOffset = offset + len;
+	    // System.err.println("startOffset = " + result.startOffset );
+	    // System.err.println("endOffset = " + result.endOffset );
+	    
+	    return result;
 	}
-
+    
 	private StringStruct getBoundedString (
 		             javax.accessibility.AccessibleText accText,
 			     org.GNOME.Accessibility.TEXT_BOUNDARY_TYPE type,
@@ -361,6 +379,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
             } else {	 
                 boundary_type = boundaryTypeConvert (type); 
                 text = accText.getAtIndex(boundary_type, offset);
+		// System.err.println ("getBoundedString: [" + text + "]");
             }
         }
         catch ( Exception e ) {
@@ -371,7 +390,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
         if (text == null) 
             stringStruct.string = "";
         else {
-            stringStruct = findOffsets (accText, text, offset);
+            stringStruct = findOffsets (accText, text, offset, false);
             stringStruct.string = text;
         }
 		//System.out.println ("text found: "+stringStruct.string);
@@ -385,6 +404,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 		// TODO: debug
         String text = "";
         try {
+	    boolean leading_ws = false;
             if (accText != null) {
                 StringStruct stringStruct = getBoundedString (accText, type, offset);
                 if (stringStruct.string != null) 
@@ -398,8 +418,14 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                     } else { 
                         text = concatenateWhitespace (accText, text, type, 
                                 offset-stringStruct.string.length());
+			if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_SENTENCE_END) {
+			    // trim trailing whitespace
+			    text = text.trim();
+			    // and add leading whitespace
+			    leading_ws = true;
+			}
                     }
-                    stringStruct = findOffsets (accText, text, offset);
+                    stringStruct = findOffsets (accText, text, offset, leading_ws);
                 }
                 startOffset.value = stringStruct.startOffset;
                 endOffset.value = stringStruct.endOffset;
@@ -423,10 +449,12 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 //                            + startOffset.value + ", "
 //                            + endOffset.value + ") " );
         try {
+	    boolean leading_ws = false;
             if (accText != null) {
                 StringStruct stringStruct = getBoundedString (accText, type, offset);
                 if (stringStruct.string != null) 
                     text = stringStruct.string;
+		//System.err.println ("getTextAtOffset [" + offset + "]: " + text);
                 if (type != TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_LINE_START
                        && type != TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_LINE_END
                        && type != TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_CHAR) {
@@ -434,8 +462,14 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                         text = concatenateText (accText, text, type, offset);
                     } else { 
                         text = concatenateWhitespace (accText, text, type, offset);
+			if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_SENTENCE_END) {
+			    // trim trailing whitespace
+			    text = text.trim();
+			    // and add leading whitespace
+			    leading_ws = true;
+			}
                     }
-                    stringStruct = findOffsets (accText, text, offset);
+                    stringStruct = findOffsets (accText, text, offset, leading_ws);
                 }
                 startOffset.value = stringStruct.startOffset;
                 endOffset.value = stringStruct.endOffset;
@@ -455,6 +489,7 @@ public class TextImpl extends UnknownImpl implements TextOperations {
 		// TODO: debug
 	    String text = "";
         try {
+	    boolean leading_ws = false;
             if (accText != null) {
                 StringStruct stringStruct = getBoundedString (accText, type, offset);
                 if (stringStruct.string != null) 
@@ -468,8 +503,14 @@ public class TextImpl extends UnknownImpl implements TextOperations {
                     } else { 
                         text = concatenateWhitespace (accText, text, type,
                               offset+stringStruct.string.length());
+			if (type == TEXT_BOUNDARY_TYPE.TEXT_BOUNDARY_SENTENCE_END) {
+			    // trim trailing whitespace
+			    text = text.trim();
+			    // and add leading whitespace
+			    leading_ws = true;			
+			}
                     }
-                    stringStruct = findOffsets (accText, text, offset);
+                    stringStruct = findOffsets (accText, text, offset, leading_ws);
                 }
                 startOffset.value = stringStruct.startOffset;
                 endOffset.value = stringStruct.endOffset;
